package lab;

import lab.data.Commit;
import lab.data.CommitToPatch;
import lab.data.Patch;
import lab.repository.CommitRepository;
import lab.repository.CommitToPatchRepository;
import lab.repository.PatchRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;

@Service
@ConditionalOnProperty(name = "job.enabled", havingValue = "CommitToPatchJob")
public class CommitToPatchJob implements CommandLineRunner {

    @Autowired
    private CommitRepository commitRepository;

    @Autowired
    private CommitToPatchRepository commitToPatchRepository;

    @Autowired
    private PatchRepository patchRepository;

    @Autowired
    private ExecutorService taskExecutor;

    private long startTime = System.currentTimeMillis();

    private long lastPrintTime = 0L;

    private int totalCommit = 0;

    private Set<String> successCommits = ConcurrentHashMap.newKeySet();

    private Set<String> failCommits = ConcurrentHashMap.newKeySet();

    private void successOne(String commitHash){
        successCommits.add(commitHash);
    }

    private void failOne(String commitHash){
        failCommits.add(commitHash);
    }

    private boolean printInfo(){
        long timestamp = System.currentTimeMillis();
        if(timestamp - lastPrintTime > 30000L){
            int successCount = successCommits.size();
            int failCount = failCommits.size();
            lastPrintTime = timestamp;
            System.out.println((timestamp - startTime) + ": " +
                    successCount + "/" + totalCommit + ", " +
                    failCount + "/" + totalCommit);
            return successCount + failCount == totalCommit;
        }
        return false;
    }

    @Override
    public void run(String... strings) throws Exception {
        List<String> commitHashList = commitRepository.findAllCommitHash();
        totalCommit = commitHashList.size();
        for (String commitHash : commitHashList) {
            printInfo();
            taskExecutor.execute(() -> processCommit(commitHash));
        }
        while(true){
            if(printInfo()) break;
            try {
                Thread.sleep(30000L);
            } catch (InterruptedException e) {
                System.out.println("sleep interrupted");
                break;
            }
        }
    }

    private void processCommit(String commitHash){
        CommitToPatch commitToPatch = commitToPatchRepository.findOne(commitHash);
        if(commitToPatch != null && commitToPatch.getPatchworkId() > 0){
            successOne(commitHash);
            return;
        }
        if(commitToPatch == null) {
            commitToPatch = new CommitToPatch()
                    .setCommitHash(commitHash);
        }
        Commit commit = commitRepository.findOne(commitHash);
        List<Patch> patchCandidates = patchRepository.findBySendTime(commit.getAuthorDate());
        Patch patch = null;
        if(patchCandidates.size() == 1){
            patch = patchCandidates.get(0);
        } else {
            for (Patch patchCandidate : patchCandidates) {
                if(subjectMatch(commit, patchCandidate)) {
                    patch = patchCandidate;
                    break;
                }
            }
            if(patch == null) {
                String query = processFullTextQuery(commit.getSubject());
                try {
                    patchCandidates = patchRepository.findBySubjectLike(query);
                } catch (Exception e){
                    System.out.println(commit.getSubject());
                    System.out.println(query);
                    throw e;
                }
                for (Patch patchCandidate : patchCandidates) {
                    if(subjectMatch(commit, patchCandidate)) {
                        patch = patchCandidate;
                        break;
                    }
                }
            }
        }
        if(patch == null){
            failOne(commitHash);
            commitToPatch.setPatchworkId(0L)
                    .setAuthorDateMatch(false)
                    .setSubjectMatch(false)
                    .setAuthorMatch(false);
        }else{
            successOne(commitHash);
            commitToPatch.setPatchworkId(patch.getPatchworkId())
                    .setAuthorDateMatch(authorDateMatch(commit, patch))
                    .setSubjectMatch(subjectMatch(commit, patch))
                    .setAuthorMatch(authorMatch(commit, patch));
        }
        commitToPatchRepository.save(commitToPatch);
    }

    public static String processFullTextQuery(String subject){
        String[] words = subject.split(" ");
        StringBuilder sb = new StringBuilder();
        for (String word : words) {
            boolean notValid = false;
            for (char c : word.toCharArray()) {
                if(c != '_' && !('0' <= c && c <= '9') && !('a' <= c && c <= 'z') && !('A' <= c && c <= 'Z')){
                    notValid = true;
                    break;
                }
            }
            if(word.trim().isEmpty()) notValid = true;
            if(notValid) continue;
            sb.append("+");
            sb.append(word);
            sb.append(" ");
        }
        return sb.toString();
    }

    public static boolean authorDateMatch(Commit commit, Patch patch){
        return commit.getAuthorDate() == patch.getSendTime();
    }

    public static boolean subjectMatch(Commit commit, Patch patch){
        return patch.getSubject() != null && patch.getSubject().contains(commit.getSubject());
    }

    public static boolean authorMatch(Commit commit, Patch patch){
        return commit.getAuthorName().equals(patch.getAuthorName()) &&
                commit.getAuthorEmail().equals(patch.getAuthorEmail());
    }
}
