package lab;

import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

@Service
@ConditionalOnProperty(name = "job.enabled", havingValue = "DownloadIndexJob")
public class DownloadIndexJob implements CommandLineRunner {

    public long startTime = System.currentTimeMillis();

    public Set<Long> patchIdSet = new HashSet<>();

    public BufferedWriter writer;

    public Queue<Integer> successPages = new ConcurrentLinkedQueue<>();

    public Queue<Integer> failPages = new ConcurrentLinkedQueue<>();

    @Autowired
    public ExecutorService executor;

    public HttpClient httpClient = HttpClients.custom()
            .setMaxConnTotal(100)
            .setMaxConnPerRoute(100)
            .build();

    public void init(File patchIds) throws IOException {
        if(patchIds.exists()) {
            BufferedReader reader = new BufferedReader(new FileReader(patchIds));
            String line;
            while ((line = reader.readLine()) != null) {
                if(!"".equals(line)) {
                    patchIdSet.add(Long.parseLong(line));
                }
            }
            reader.close();
        }
        writer = new BufferedWriter(new FileWriter(patchIds, true));
        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            synchronized (writer) {
                executor.shutdownNow();
                try {
                    writer.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            printMessage();
        }));
    }

    public void printMessage(){
        System.out.println("Success page: " + new TreeSet<>(successPages));
        System.out.println("Fail page: " + new TreeSet<>(failPages));
        System.out.println("Used time: " + (System.currentTimeMillis() - startTime) / 1000);
    }

    @Override
    public void run(String... args) throws Exception {
        init(new File("patchIds"));
//        List<Integer> pages = Arrays.asList(1, 6116, 6120, 6122, 6138, 6147, 6666);
//        for(int page : pages){
        for(int page = 1; page <= 6666; ++page){
            int page0 = page;
            String pageUrl = "https://patchwork.kernel.org/project/LKML/list/?state=*&archive=both&page="+page;
            executor.execute(() -> {
                try {
                    extractPatchIds(pageUrl);
                    successPages.add(page0);
                } catch (Throwable e) {
                    System.out.println("error: " + e.getMessage() + ", page: " + page0);
                    failPages.add(page0);
                }
            });
        }
        executor.shutdown();
        while(true) {
            printMessage();
            try {
                if(executor.awaitTermination(15, TimeUnit.SECONDS)){
                    break;
                }
            } catch (InterruptedException e) {

            }
        }
    }

    public void extractPatchIds(String indexPageUrl) throws IOException {
        Document doc = Jsoup.parse(httpGetContent(indexPageUrl));
//        Document doc = Jsoup.connect(indexPageUrl).get();
        Element patchTable = doc.getElementById("patchlist");
        Elements patchHrefs = patchTable.select("tbody > tr > td > a[href^=\"/patch/\"]");
        Set<Long> patchIds = new HashSet<>();
        patchHrefs.forEach(patchHref -> {
            String href = patchHref.attr("href");
            patchIds.add(Long.parseLong(href.split("/")[2]));
        });
        dumpPatchIds(patchIds);
    }

    public void dumpPatchIds(Set<Long> patchIds) throws IOException {
        synchronized (writer) {
            patchIds.forEach(patchId -> {
                if (!patchIdSet.contains(patchId)) {
                    try {
                        patchIdSet.add(patchId);
                        writer.write(String.valueOf(patchId));
                        writer.newLine();
                    } catch (IOException e) {
                        throw new RuntimeException(e.getMessage(), e);
                    }
                }
            });
        }
    }

    public String httpGetContent(String url) throws IOException {
        HttpResponse response = httpClient.execute(new HttpGet(url));
        int statusCode = response.getStatusLine().getStatusCode();
        if(statusCode < 200 || statusCode >= 300)
            throw new RuntimeException("status Code: " + statusCode + ", reason: " + response.getStatusLine().getReasonPhrase());
        return EntityUtils.toString(response.getEntity());
    }
}
