package lab.service;

import lab.data.Patch;
import lab.data.PatchComment;
import lab.repository.PatchCommentRepository;
import lab.repository.PatchRepository;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.*;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

@Service
public class PatchService {

    @Autowired
    private PatchRepository patchRepository;

    @Autowired
    private PatchCommentRepository patchCommentRepository;

    @Value("${processPatch.patchDir}")
    private String patchDir = "";

    @Transactional
    public void processPatch(long patchworkId) throws Exception {
        File mboxFile = new File(patchDir + patchworkId + "\\mbox.patch");
        File webFile = new File(patchDir + patchworkId + "\\web.html");
        Patch patch = new Patch().setPatchworkId(patchworkId);
        try(BufferedReader reader = new BufferedReader(new FileReader(mboxFile))){
            StringBuilder mboxContent = new StringBuilder((int)mboxFile.length() + 1);
            String line;
            StringBuilder lineBuilder = new StringBuilder();
            while((line = reader.readLine()) != null){
                mboxContent.append(line);
                mboxContent.append("\n");
                if(line.startsWith("\t") || line.startsWith(" ")){
                    lineBuilder.append(" ");
                    lineBuilder.append(line.substring(1));
                }else{
                    processMboxLine(patch, lineBuilder.toString());
                    lineBuilder = new StringBuilder(line);
                }
            }
            processMboxLine(patch, lineBuilder.toString());
            patch.setMbox(mboxContent.toString());
        }
        Document webContent = Jsoup.parse(webFile, "utf-8");
        processPatchFromWeb(patch, webContent);
        List<PatchComment> patchComments = getPatchCommentsFromWeb(patchworkId, webContent);
        patchCommentRepository.save(patchComments);
        patchRepository.save(patch);
    }

    private void processMboxLine(Patch patch, String line) {
        try {
            int firstSpace = line.indexOf(' ');
            if (firstSpace <= 0) return;
            if (line.charAt(firstSpace - 1) != ':') return;
            String headerName = line.substring(0, firstSpace - 1);
            String value = line.substring(firstSpace + 1).trim();
            int valueLen = value.length();
            if(valueLen == 0) return;
            switch (headerName) {
                case "Subject":
                    if(patch.getSubject() != null) break;
                    patch.setSubject(value);
                    break;
                case "From":
                    if(patch.getAuthorEmail() != null) break;
                    int lastSpace = value.lastIndexOf(' ');
                    if (lastSpace <= 0) {
                        patch.setAuthorEmail(value);
                    } else {
                        patch.setAuthorName(value.substring(0, lastSpace));
                        patch.setAuthorEmail(value.substring(lastSpace + 2, valueLen - 1));
                    }
                    break;
                case "Message-Id":
                    if(patch.getMessageId() != null) break;
                    patch.setMessageId(value.substring(1, valueLen - 1));
                    break;
            }
        } catch (Exception e){
            System.out.println(line);
            e.printStackTrace();
            throw e;
        }
    }

    private static long parseDate(String[] datePatternList, String value, boolean isUtc) throws ParseException {
        for (String pattern : datePatternList) {
            try {
                SimpleDateFormat dateFormat = new SimpleDateFormat(pattern, Locale.ENGLISH);
                if(isUtc) dateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
                return dateFormat.parse(value).getTime();
            } catch (ParseException e) {

            }
        }
        throw new ParseException("date parse fail: " + value, 0);
    }

    private static String[] authorDatePatternList = new String[]{
            "EEE, dd MMM yyyy HH:mm:ss Z",
            "dd MMM yyyy HH:mm:ss Z",
            "EEE, dd MMM yyyy HH:mm Z",
            "dd MMM yyyy HH:mm Z",
            "EEE, MMM dd HH:mm:ss yyyy Z",
            "EEE, MMM dd HH:mm:ss z yyyy"
    };

    private void processPatchFromWeb(Patch patch, Document webContent) throws ParseException {
        Element headers = webContent.getElementById("patchheaders");
        String[] headerList = headers.text().split("\n");
        for (String line : headerList) {
            try {
                int firstSpace = line.indexOf(' ');
                if (firstSpace <= 0) continue;
                if (line.charAt(firstSpace - 1) != ':') continue;
                String headerName = line.substring(0, firstSpace - 1);
                String value = line.substring(firstSpace + 1).trim();
                int valueLen = value.length();
                if(valueLen == 0) continue;
                switch (headerName) {
                    case "Date":
                        patch.setSendTime(parseDate(authorDatePatternList, value, false));
                        patch.setSendTimeRaw(value);
                        break;
                    case "In-Reply-To":
                        patch.setInReplyTo(value.substring(1, valueLen - 1));
                        break;
                    case "References":
                        patch.setReferenceList(value.substring(1, valueLen - 1));
                        break;
                }
            }catch (Exception e){
                System.out.println(line);
                e.printStackTrace();
                throw e;
            }
        }
    }

    private static String[] commentDatePatternList = new String[]{
            "- MMM dd, yyyy, hh:mm a",
            "- MMM dd, yyyy, hh a"
    };

    public static void main(String[] args) throws ParseException {
        System.out.println(parseDate(commentDatePatternList, "- Sept. 8, 2017, 6:15 PM", true));
    }

    private List<PatchComment> getPatchCommentsFromWeb(long patchworkId, Document webContent) throws ParseException {
        Elements commentElements = webContent.select("div.comment");
        List<PatchComment> result = new ArrayList<>(commentElements.size());
        for (Element commentElement : commentElements) {
            PatchComment comment = new PatchComment().setPatchworkId(patchworkId);

            Element meta = commentElement.selectFirst("div.meta");
            String timeStringRaw = meta.ownText().trim();
            String timeString = timeStringRaw
                    .replace(".", "")
                    .replace("a.m.", "AM")
                    .replace("p.m.", "PM")
                    .replace("noon", "12 PM")
                    .replace("midnight", "12 AM")
                    .replace("Sept", "Sep");
            comment.setCommentTime(parseDate(commentDatePatternList, timeString, true))
                    .setCommentTimeRaw(timeStringRaw);

            Element author = meta.selectFirst("a");
            comment.setAuthorName(author.ownText().trim());

            String authorIdUri = author.attr("href");
            if(authorIdUri == null || !authorIdUri.startsWith("/project/LKML/list/?submitter="))
                throw new RuntimeException("authorId not found");
            comment.setPwAuthorId(Long.parseLong(authorIdUri.substring("/project/LKML/list/?submitter=".length())));

            comment.setContent(commentElement.selectFirst("pre.content").text());

            result.add(comment);
        }
        return result;
    }
}
